from __future__ import absolute_import, print_function, unicode_literals

import threading


class PollThread(threading.Thread):

    def __init__(self, operation, poll_interval=1, poll_retries=10):
        super(PollThread, self).__init__()
        self.daemon = True
        self.operation = operation
        self.id = operation.id
        self.poll_interval = poll_interval
        self.poll_retries = poll_retries

    def poll(self):
        raise Exception("PollThread.poll() not implemented")

    def is_done(self, operation):
        raise Exception("PollThread.is_done(operation) not implemented")

    def complete(self, operation):
        raise Exception("PollThread.complete(operation) not implemented")

    def run(self):
        raise Exception("PollThread.run() not implemented")



from __future__ import print_function, unicode_literals

import glob
import os
import shutil
import subprocess
import sys
import time

from .funnel_test_util import SimpleServerTest, popen


class TestConformance(SimpleServerTest):

    def test_conformance(self):
        cwl_testdir = os.path.join(self.testdir, "schemas/v1.0")
        ctest_def = os.path.join(cwl_testdir, "conformance_test_v1.0.yaml")
        tool_entry = os.path.join(self.rootprojectdir, "cwl-tes")

        cmd = [
            "cwltest", "--test", ctest_def, "--basedir", self.tmpdir,
            "--tool", tool_entry, "-j", "20", "--",
            "--tes=http://localhost:8000"
        ]

        testlog = os.path.join(
            self.tmpdir,
            "conformance_test_results_%s.txt" % (time.strftime("%m%d%Y"))
        )
        print("RUNNNING:", " ".join(cmd))
        p = popen(
            cmd,
            cwd=os.path.join(self.testdir, "schemas/v1.0"),
            stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE,
            bufsize=1
        )
        with p.stdout, open(testlog, 'ab') as fh:
            for line in iter(p.stdout.readline, b''):
                print(line.decode('utf8').strip("\n"), file=sys.stderr)
                fh.write(line)
        p.wait()

        ctest_dirs = glob.glob(cwl_testdir + "[a-zA-Z0-9_]*")
        cleanup_tmpdirs(ctest_dirs)

        assert p.returncode == 0


def cleanup_tmpdirs(*args):
    for d in args:
        if isinstance(d, list):
            for sd in d:
                shutil.rmtree(sd, True)
        else:
            shutil.rmtree(d, True)



from __future__ import absolute_import, print_function, unicode_literals

import logging
import os
import shutil
import tes
import time

from cwltool.draft2tool import CommandLineTool
from cwltool.errors import WorkflowException, UnsupportedRequirement
from cwltool.pathmapper import PathMapper
from cwltool.stdfsaccess import StdFsAccess
from cwltool.workflow import defaultMakeTool
from pprint import pformat
from schema_salad.ref_resolver import file_uri

from cwl_tes.pipeline import Pipeline, PipelineJob
from cwl_tes.poll import PollThread


log = logging.getLogger("tes-backend")


class TESPipeline(Pipeline):

    def __init__(self, url, kwargs):
        super(TESPipeline, self).__init__()
        self.kwargs = kwargs
        self.service = tes.HTTPClient(url)
        if kwargs.get("basedir") is not None:
            self.basedir = kwargs.get("basedir")
        else:
            self.basedir = os.getcwd()
        self.fs_access = StdFsAccess(self.basedir)

    def make_exec_tool(self, spec, **kwargs):
        return TESPipelineTool(spec, self, fs_access=self.fs_access, **kwargs)

    def make_tool(self, spec, **kwargs):
        if "class" in spec and spec["class"] == "CommandLineTool":
            return self.make_exec_tool(spec, **kwargs)
        else:
            return defaultMakeTool(spec, **kwargs)


class TESPipelineTool(CommandLineTool):

    def __init__(self, spec, pipeline, fs_access, **kwargs):
        super(TESPipelineTool, self).__init__(spec, **kwargs)
        self.spec = spec
        self.pipeline = pipeline
        self.fs_access = fs_access

    def makeJobRunner(self, use_container=True):
        return TESPipelineJob(self.spec, self.pipeline, self.fs_access)

    def makePathMapper(self, reffiles, stagedir, **kwargs):
        return PathMapper(reffiles, kwargs["basedir"], stagedir)


class TESPipelineJob(PipelineJob):

    def __init__(self, spec, pipeline, fs_access):
        super(TESPipelineJob, self).__init__(spec, pipeline)
        self.outputs = None
        self.docker_workdir = "/var/spool/cwl"
        self.fs_access = fs_access
        self.inplace_update = False

    def create_input_parameter(self, name, d):
        if "contents" in d:
            return tes.TaskParameter(
                name=name,
                description="cwl_input:%s" % (name),
                path=d["path"],
                contents=d["contents"],
                type=d["class"].upper()
            )
        else:
            return tes.TaskParameter(
                name=name,
                description="cwl_input:%s" % (name),
                url=d["location"],
                path=d["path"],
                type=d["class"].upper()
            )

    def parse_job_order(self, k, v, inputs):
        if isinstance(v, dict):
            if all([i in v for i in ["location", "path", "class"]]):
                inputs.append(self.create_input_parameter(k, v))

                if "secondaryFiles" in v:
                    for f in v["secondaryFiles"]:
                        self.parse_job_order(f["basename"], f, inputs)

            else:
                for sk, sv in v.items():
                    if isinstance(sv, dict):
                        self.parse_job_order(sk, sv, inputs)

                    else:
                        break

        elif isinstance(v, list):
            for i in range(len(v)):
                if isinstance(v[i], dict):
                    self.parse_job_order("%s[%s]" % (k, i), v[i], inputs)

                else:
                    break

        return inputs

    def parse_listing(self, listing, inputs):
        for item in listing:

            if "writable" in item:
                raise UnsupportedRequirement(
                    "The TES spec does not allow for writable inputs"
                    )

            if "contents" in item:
                loc = self.fs_access.join(self.tmpdir, item["basename"])
                with self.fs_access.open(loc, "wb") as gen:
                    gen.write(item["contents"])
            else:
                loc = item["location"]

            parameter = tes.TaskParameter(
                name=item["basename"],
                description="InitialWorkDirRequirement:cwl_input:%s" % (
                    item["basename"]
                ),
                url=file_uri(loc),
                path=self.fs_access.join(
                    self.docker_workdir, item["basename"]
                ),
                type=item["class"].upper()
                )
            inputs.append(parameter)

        return inputs

    def collect_input_parameters(self):
        inputs = []

        # find all primary and secondary input files
        for k, v in self.joborder.items():
            self.parse_job_order(k, v, inputs)

        # manage InitialWorkDirRequirement
        self.parse_listing(self.generatefiles["listing"], inputs)

        return inputs

    def create_task_msg(self):
        input_parameters = self.collect_input_parameters()
        output_parameters = []

        if self.stdout is not None:
            parameter = tes.TaskParameter(
                name="stdout",
                url=self.output2url(self.stdout),
                path=self.output2path(self.stdout)
            )
            output_parameters.append(parameter)

        if self.stderr is not None:
            parameter = tes.TaskParameter(
               name="stderr",
               url=self.output2url(self.stderr),
               path=self.output2path(self.stderr)
            )
            output_parameters.append(parameter)

        output_parameters.append(
            tes.TaskParameter(
                name="workdir",
                url=self.output2url(""),
                path=self.docker_workdir,
                type="DIRECTORY"
            )
        )

        container = self.find_docker_requirement()

        cpus = None
        ram = None
        disk = None
        for i in self.requirements:
            if i.get("class", "NA") == "ResourceRequirement":
                cpus = i.get("coresMin", i.get("coresMax", None))
                ram = i.get("ramMin", i.get("ramMax", None))
                ram = ram / 953.674 if ram is not None else None
                disk = i.get("outdirMin", i.get("outdirMax", None))
                disk = disk / 953.674 if disk is not None else None
            elif i.get("class", "NA") == "DockerRequirement":
                if i.get("dockerOutputDirectory", None) is not None:
                    output_parameters.append(
                        tes.TaskParameter(
                            name="dockerOutputDirectory",
                            url=self.output2url(""),
                            path=i.get("dockerOutputDirectory"),
                            type="DIRECTORY"
                        )
                    )

        create_body = tes.Task(
            name=self.name,
            description=self.spec.get("doc", ""),
            executors=[
                tes.Executor(
                    cmd=self.command_line,
                    image_name=container,
                    workdir=self.docker_workdir,
                    stdout=self.output2path(self.stdout),
                    stderr=self.output2path(self.stderr),
                    stdin=self.stdin,
                    environ=self.environment
                )
            ],
            inputs=input_parameters,
            outputs=output_parameters,
            resources=tes.Resources(
                cpu_cores=cpus,
                ram_gb=ram,
                size_gb=disk
            ),
            tags={"CWLDocumentId": self.spec.get("id")}
        )

        return create_body

    def run(self, pull_image=True, rm_container=True, rm_tmpdir=True,
            move_outputs="move", **kwargs):
        # useful for debugging
        log.debug(
            "[job %s] self.__dict__ in run() ----------------------" %
            (self.name)
        )
        log.debug(pformat(self.__dict__))

        task = self.create_task_msg()

        log.info(
            "[job %s] CREATED TASK MSG----------------------" %
            (self.name)
        )
        log.info(pformat(task))

        try:
            task_id = self.pipeline.service.create_task(task)
            log.info(
                "[job %s] SUBMITTED TASK ----------------------" %
                (self.name)
            )
            log.info("[job %s] task id: %s " % (self.name, task_id))
            operation = self.pipeline.service.get_task(task_id, "MINIMAL")
        except Exception as e:
            log.error(
                "[job %s] Failed to submit task to TES service:\n%s" %
                (self.name, e)
            )
            return WorkflowException(e)

        def callback(operation):
            try:
                outputs = self.collect_outputs(self.outdir)
                cleaned_outputs = {}
                for k, v in outputs.items():
                    if isinstance(k, bytes):
                        k = k.decode("utf8")
                    if isinstance(v, bytes):
                        v = v.decode("utf8")
                    cleaned_outputs[k] = v
                self.outputs = cleaned_outputs
                self.output_callback(self.outputs, "success")
            except WorkflowException as e:
                log.error("[job %s] job error:\n%s" % (self.name, e))
                self.output_callback({}, "permanentFail")
            except Exception as e:
                log.error("[job %s] job error:\n%s" % (self.name, e))
                self.output_callback({}, "permanentFail")
            finally:
                if self.outputs is not None:
                    log.info(
                        "[job %s] OUTPUTS ------------------" %
                        (self.name)
                    )
                    log.info(pformat(self.outputs))
                self.cleanup(rm_tmpdir)

        poll = TESPipelinePoll(
            jobname=self.name,
            service=self.pipeline.service,
            operation=operation,
            callback=callback
        )

        self.pipeline.add_thread(poll)
        poll.start()

    def cleanup(self, rm_tmpdir):
        log.debug(
            "[job %s] STARTING CLEAN UP ------------------" %
            (self.name)
        )
        if self.stagedir and os.path.exists(self.stagedir):
            log.debug(
                "[job %s] Removing input staging directory %s" %
                (self.name, self.stagedir)
            )
            shutil.rmtree(self.stagedir, True)

        if rm_tmpdir:
            log.debug(
                "[job %s] Removing temporary directory %s" %
                (self.name, self.tmpdir)
            )
            shutil.rmtree(self.tmpdir, True)

    def output2url(self, path):
        if path is not None:
            return file_uri(
                self.fs_access.join(self.outdir, os.path.basename(path))
            )
        return None

    def output2path(self, path):
        if path is not None:
            return self.fs_access.join(self.docker_workdir, path)
        return None


class TESPipelinePoll(PollThread):

    def __init__(self, jobname, service, operation, callback):
        super(TESPipelinePoll, self).__init__(operation)
        self.name = jobname
        self.service = service
        self.callback = callback

    def run(self):
        while not self.is_done(self.operation):
            time.sleep(self.poll_interval)
            # slow down polling over time till it hits a max
            # if self.poll_interval < 30:
            #     self.poll_interval += 1
            log.debug(
                "[job %s] POLLING %s" %
                (self.name, pformat(self.id))
            )
            try:
                self.operation = self.poll()
            except Exception as e:
                log.error("[job %s] POLLING ERROR %s" % (self.name, e))
                if self.poll_retries > 0:
                    self.poll_retries -= 1
                    continue
                else:
                    log.error("[job %s] MAX POLLING RETRIES EXCEEDED" %
                              (self.name))
                    break

        self.complete(self.operation)

    def poll(self):
        return self.service.get_task(self.id, "MINIMAL")

    def is_done(self, operation):
        terminal_states = ["COMPLETE", "CANCELED", "ERROR", "SYSTEM_ERROR"]
        if operation.state in terminal_states:
            log.info(
                "[job %s] FINAL JOB STATE: %s ------------------" %
                (self.name, operation.state)
            )
            if operation.state != "COMPLETE":
                log.error(
                    "[job %s] task id: %s" % (self.name, self.id)
                )
                log.error(
                    "[job %s] logs: %s" %
                    (
                        self.name,
                        self.service.get_task(self.id, "FULL").logs
                    )

                )
            return True
        return False

    def complete(self, operation):
        self.callback(operation)



from __future__ import absolute_import, print_function, unicode_literals

import logging
import os
import tempfile
import time

from builtins import str
from cwltool.errors import WorkflowException
from cwltool.process import cleanIntermediate, relocateOutputs
from cwltool.mutation import MutationManager

log = logging.getLogger("tes-backend")


class Pipeline(object):

    def __init__(self):
        self.threads = []

    def executor(self, tool, job_order, **kwargs):
        final_output = []
        final_status = []

        def output_callback(out, processStatus):
            final_status.append(processStatus)
            final_output.append(out)

        if "basedir" not in kwargs:
            raise WorkflowException("Must provide 'basedir' in kwargs")

        output_dirs = set()

        if kwargs.get("outdir"):
            finaloutdir = os.path.abspath(kwargs.get("outdir"))
        else:
            finaloutdir = None

        if kwargs.get("tmp_outdir_prefix"):
            kwargs["outdir"] = tempfile.mkdtemp(
                prefix=kwargs["tmp_outdir_prefix"]
            )
        else:
            kwargs["outdir"] = tempfile.mkdtemp()

        output_dirs.add(kwargs["outdir"])
        kwargs["mutation_manager"] = MutationManager()

        jobReqs = None
        if "cwl:requirements" in job_order:
            jobReqs = job_order["cwl:requirements"]
        elif ("cwl:defaults" in tool.metadata and
              "cwl:requirements" in tool.metadata["cwl:defaults"]):
            jobReqs = tool.metadata["cwl:defaults"]["cwl:requirements"]
        if jobReqs:
            for req in jobReqs:
                tool.requirements.append(req)

        if kwargs.get("default_container"):
            tool.requirements.insert(0, {
                "class": "DockerRequirement",
                "dockerPull": kwargs["default_container"]
            })

        jobs = tool.job(job_order, output_callback, **kwargs)
        try:
            for runnable in jobs:
                if runnable:
                    builder = kwargs.get("builder", None)
                    if builder is not None:
                        runnable.builder = builder
                    if runnable.outdir:
                        output_dirs.add(runnable.outdir)
                    runnable.run(**kwargs)
                else:
                    # log.error(
                    #     "Workflow cannot make any more progress"
                    # )
                    # break
                    time.sleep(1)

        except WorkflowException as e:
            raise e
        except Exception as e:
            log.error("Got exception")
            raise WorkflowException(str(e))

        # wait for all processes to finish
        self.wait()

        if final_output and final_output[0] and finaloutdir:
            final_output[0] = relocateOutputs(
                final_output[0], finaloutdir,
                output_dirs, kwargs.get("move_outputs"),
                kwargs["make_fs_access"](""))

        if kwargs.get("rm_tmpdir"):
            cleanIntermediate(output_dirs)

        if final_output and final_status:
            return (final_output[0], final_status[0])
        else:
            return (None, "permanentFail")

    def make_exec_tool(self, spec, **kwargs):
        raise Exception("Pipeline.make_exec_tool() not implemented")

    def make_tool(self, spec, **kwargs):
        raise Exception("Pipeline.make_tool() not implemented")

    def add_thread(self, thread):
        self.threads.append(thread)

    def wait(self):
        while True:
            if all([not t.is_alive() for t in self.threads]):
                break
        for t in self.threads:
            t.join()


class PipelineJob(object):

    def __init__(self, spec, pipeline):
        self.spec = spec
        self.pipeline = pipeline
        self.running = False

    def find_docker_requirement(self):
        default = "python:2.7"
        container = default
        if self.pipeline.kwargs["default_container"]:
            container = self.pipeline.kwargs["default_container"]

        reqs = self.spec.get("requirements", []) + self.spec.get("hints", [])
        for i in reqs:
            if i.get("class", "NA") == "DockerRequirement":
                container = i.get(
                    "dockerPull",
                    i.get("dockerImageId", default)
                )
        return container

    def run(self, pull_image=True, rm_container=True, rm_tmpdir=True,
            move_outputs="move", **kwargs):
        raise Exception("PipelineJob.run() not implemented")



from __future__ import print_function

import logging
import os
import signal
import subprocess
import sys
import tempfile
import time
import unittest
import yaml


def cmd_exists(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return True
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return True

    return False


def popen(*args, **kwargs):
    kwargs['preexec_fn'] = os.setsid
    return subprocess.Popen(*args, **kwargs)


def kill(p):
    try:
        os.killpg(os.getpgid(p.pid), signal.SIGTERM)
        p.wait()
    except OSError:
        pass


def temp_config(dir, config):
    configFile = tempfile.NamedTemporaryFile(dir=dir, mode='w', delete=False)
    yaml.dump(config, configFile)
    return configFile


def config_seconds(sec):
    # The funnel config is currently parsed as nanoseconds
    # this helper makes that manageale
    return int(sec * 1000000000)


class SimpleServerTest(unittest.TestCase):

    def setUp(self):
        self.tmpdir = None
        self.task_server = None
        self.addCleanup(self.cleanup)

        if not cmd_exists("funnel"):
            print(
                "-bash: funnel: command not found\n",
                "see https://ohsu-comp-bio.github.io/funnel/install/",
                "for instuctions on how to install",
                file=sys.stdout
            )
            raise RuntimeError

        self.rootprojectdir = os.path.dirname(os.path.dirname(
            os.path.realpath(__file__)
        ))

        self.testdir = os.path.join(self.rootprojectdir, "tests")
        if not os.path.exists(os.path.join(self.testdir, "test_tmp")):
            os.mkdir(os.path.join(self.testdir, "test_tmp"))
        self.tmpdir = tempfile.mkdtemp(
            dir=os.path.join(self.testdir, "test_tmp"),
            prefix="conformance_test_v1.0_"
        )
        os.environ['TMPDIR'] = self.tmpdir

        f, db_path = tempfile.mkstemp(dir=self.tmpdir, prefix="tes_task_db.")
        os.close(f)
        funnel_work_dir = os.path.join(self.tmpdir, "funnel-work-dir")
        os.mkdir(funnel_work_dir)
        logFile = os.path.join(self.tmpdir, "funnel_log.txt")

        # Build server config file (YAML)
        rate = config_seconds(0.05)
        configFile = temp_config(dir=self.tmpdir, config={
            "HostName": "localhost",
            "HTTPPort": "8000",
            "RPCPort": "9090",
            "DBPath": db_path,
            "WorkDir": funnel_work_dir,
            "Storage": {
                "Local": {
                    "AllowedDirs": [self.testdir]
                }
            },
            "Logger": {
                "Level": "debug",
                "OutputFile": logFile,
            },
            "Worker": {
                "WorkDir": funnel_work_dir,
                "Logger": {
                    "Level": "debug",
                    "OutputFile": logFile,
                },
                "Timeout": -1,
                "StatusPollRate": rate,
                "LogUpdateRate": rate,
                "NewJobPollRate": rate,
                "UpdateRate": rate,
                "TrackerRate": rate
            },
            "ScheduleRate": rate,
        })

        # Start server
        cmd = ["funnel", "server", "--config", configFile.name]
        logging.info("Running %s" % (" ".join(cmd)))
        self.task_server = popen(cmd)
        signal.signal(signal.SIGINT, self.cleanup)
        time.sleep(1)

    # We're using this instead of tearDown because python doesn't call tearDown
    # if setUp fails. Since our setUp is complex, that means things don't get
    # properly cleaned up (e.g. processes are orphaned).
    def cleanup(self, *args):
        if self.task_server is not None:
            kill(self.task_server)



from __future__ import absolute_import, print_function, unicode_literals

import cwltool.main
import pkg_resources
import signal
import sys
import logging

from cwl_tes.tes import TESPipeline
from cwl_tes.__init__ import __version__


log = logging.getLogger("tes-backend")
log.setLevel(logging.INFO)
console = logging.StreamHandler()
# formatter = logging.Formatter("[%(asctime)s]\t[%(levelname)s]\t%(message)s")
# console.setFormatter(formatter)
log.addHandler(console)


def versionstring():
    pkg = pkg_resources.require("cwltool")
    if pkg:
        cwltool_ver = pkg[0].version
    else:
        cwltool_ver = "unknown"
    return "%s %s with cwltool %s" % (sys.argv[0], __version__, cwltool_ver)


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    parser = cwltool.main.arg_parser()
    parser = add_args(parser)
    parsed_args = parser.parse_args(args)

    if not len(args) >= 1:
        print(versionstring())
        print("CWL document required, no input file was provided")
        parser.print_usage()
        return 1

    if parsed_args.version:
        print(versionstring())
        return 0

    if parsed_args.tes is None:
        print(versionstring())
        parser.print_usage()
        print("cwl-tes: error: argument --tes is required")
        return 1

    if parsed_args.quiet:
        log.setLevel(logging.WARN)
    if parsed_args.debug:
        log.setLevel(logging.DEBUG)

    blacklist_false = ["no_container", "disable_pull", "disable_net",
                       "custom_net", "no_match_user"]
    for f in blacklist_false:
        if vars(parsed_args).get(f):
            log.warning("arg: '%s' has no effect in cwl-tes" % (f))

    blacklist_true = ["enable_pull"]
    for f in blacklist_true:
        if not vars(parsed_args).get(f):
            log.warning("arg: '%s' has no effect in cwl-tes" % (f))

    # custom
    if not parsed_args.rm_container:
        log.warning("arg: 'leave_container' has no effect in cwl-tes")

    pipeline = TESPipeline(parsed_args.tes, vars(parsed_args))

    # setup signal handler
    def signal_handler(*args):
        log.info(
            "recieved control-c signal"
        )
        log.info(
            "terminating thread(s)..."
        )
        log.warning(
            "remote TES processes %s may keep running" %
            ([t.id for t in pipeline.threads])
        )
        sys.exit(1)
    signal.signal(signal.SIGINT, signal_handler)

    return cwltool.main.main(
        args=parsed_args,
        executor=pipeline.executor,
        makeTool=pipeline.make_tool,
        versionfunc=versionstring,
        logger_handler=console
    )


def add_args(parser):
    parser.add_argument(
        "--tes",
        type=str,
        help="GA4GH TES Service URL"
    )
    return parser


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

from __future__ import print_function

import io
import os
import re

from setuptools import setup, find_packages


def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="cwl-tes",
    version=find_version("cwl_tes", "__init__.py"),
    description="Common workflow language reference implementation backended \
    by a GA4GH Task Execution Service",
    author="Adam Struck",
    author_email="strucka@ohsu.edu",
    url="https://github.com/common-workflow-language/cwl-tes",
    license="Apache 2.0",
    packages=find_packages(),
    python_requires=">=2.7, <4",
    install_requires=[
        "cwltool==1.0.20170727112954",
        "future>=0.16.0",
        "py-tes>=0.1.5",
        "requests>=2.14.2"
    ],
    entry_points={
        "console_scripts": ["cwl-tes=cwl_tes.main:main"]},
    zip_safe=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Natural Language :: English",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
)
